@@ -50,15 +50,46 @@ class MLP:
         self.accuracy_list = []
         self.valid_accuracy_list = []
 
+        self.min_valid_loss = math.inf
+        self.min_valid_cnt = 0
+        self.validation_set_early_stop = False
 
 
-    def train(self, x_train, y_train, x_valid='NULL', y_valid='NULL', lr=0.01, epochs=100, batch_size=1, verbose_frequence=1):
+
+    def train(self, x_train, y_train, x_valid='NULL', y_valid='NULL', lr=0.01, epochs=100, batch_size=1, verbose_frequence=1, methods=False):
         if len(x_train) % batch_size != 0:
             raise ValueError(f"batch_size需为训练集样本数量的约数")
+
+        if isinstance(methods, dict):
+            for method in methods:
+                if not (method == 'lr_decay' or method == 'validation_set_early_stop'):
+                    raise ValueError(f"目前methods支持lr_decay及validation_set_early_stop")
+                if method == 'lr_decay':
+                    if not methods['lr_decay'][0] == 'exp_decay':
+                        raise ValueError(f"目前lr_decay支持exp_decay")
+                    else:
+                        if not isinstance(methods['lr_decay'][1], (int, float)):
+                            raise ValueError(f"列表第二列请输入正确的整形或字符型衰减率")
+                    lr0 = lr
+                if method == 'validation_set_early_stop':
+                    if not isinstance(methods['validation_set_early_stop'][0], int):
+                        raise ValueError(f"请输入合法的验证集提前终止代数（整形）")
+                    else:
+                        if not methods['validation_set_early_stop'][1] is False:
+                            raise ValueError(f"目前只支持保存提前终止后训练的最后一代模型的参数，验证集损失最小的模型暂不支持")
+                        else:
+                            self.validation_set_early_stop = methods['validation_set_early_stop'][0]
+
         ac = getattr(activation_function, self.hidden_activation)
         ac_derivative = getattr(activation_function, self.hidden_activation + '_derivative')
         if self.target == '2_classify':
             for _ in tqdm(range(epochs)):
+                text = f'第{_ + 1}代'
+                if isinstance(methods, dict):
+                    if 'lr_decay' in methods:
+                        lr = lr0 * np.exp(-methods['lr_decay'][1] * _)
+                        text += f'学习率：{lr}，'
+
 
                 indices = list(range(len(x_train)))
                 random.shuffle(indices)
@@ -180,18 +211,35 @@ class MLP:
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
                     self.valid_accuracy_list.append(valid_cnt / len(x_valid))
                 if (_ + 1) % verbose_frequence == 0:
+                    text += f'损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%"
-                              f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%")
+                        text += f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%"
+                    print(text)
+
+                if isinstance(x_valid, list) and isinstance(y_valid, list) and self.validation_set_early_stop is not False:
+                    if self.valid_loss_values[-1] < self.min_valid_loss:
+                        self.min_valid_cnt = 0
+                        self.min_valid_loss = self.valid_loss_values[-1]
                     else:
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%")
+                        self.min_valid_cnt += 1
+
+                if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
+                    break
 
 
-
-            return self.loss_values, self.accuracy_list, self.valid_loss_values, self.valid_accuracy_list
+            if isinstance(x_valid, list) and isinstance(y_valid, list):
+                return self.loss_values, self.accuracy_list, self.valid_loss_values, self.valid_accuracy_list
+            else:
+                return self.loss_values, self.accuracy_list
 
         if self.target == 'multi_classify':
             for _ in tqdm(range(epochs)):
+                text = f'第{_ + 1}代'
+                if isinstance(methods, dict):
+                    if 'lr_decay' in methods:
+                        lr = lr0 * np.exp(-methods['lr_decay'][1] * _)
+                        text += f'学习率：{lr}，'
 
                 indices = list(range(len(x_train)))
                 random.shuffle(indices)
@@ -340,16 +388,34 @@ class MLP:
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
                     self.valid_accuracy_list.append(valid_cnt / len(x_valid))
                 if (_ + 1) % verbose_frequence == 0:
+                    text += f'损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%"
-                              f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%")
-                    else:
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，准确率：{self.accuracy_list[-1] * 100}%")
+                        text += f"，验证集损失：{self.valid_loss_values[-1]}，验证集准确率：{self.valid_accuracy_list[-1] * 100}%"
+                    print(text)
 
-            return self.loss_values, self.accuracy_list, self.valid_loss_values, self.valid_accuracy_list
+                if isinstance(x_valid, list) and isinstance(y_valid, list) and self.validation_set_early_stop is not False:
+                    if self.valid_loss_values[-1] < self.min_valid_loss:
+                        self.min_valid_cnt = 0
+                        self.min_valid_loss = self.valid_loss_values[-1]
+                    else:
+                        self.min_valid_cnt += 1
+
+                if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
+                    break
+
+            if isinstance(x_valid, list) and isinstance(y_valid, list):
+                return self.loss_values, self.accuracy_list, self.valid_loss_values, self.valid_accuracy_list
+            else:
+                return self.loss_values, self.accuracy_list
 
         if self.target == 'output_1_node_regress':
             for _ in tqdm(range(epochs)):
+                text = f'第{_ + 1}代'
+                if isinstance(methods, dict):
+                    if 'lr_decay' in methods:
+                        lr = lr0 * np.exp(-methods['lr_decay'][1] * _)
+                        text += f'学习率：{lr}，'
 
                 indices = list(range(len(x_train)))
                 random.shuffle(indices)
@@ -456,15 +522,35 @@ class MLP:
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
 
                 if (_ + 1) % verbose_frequence == 0:
+                    text += f'损失：{self.loss_values[-1]}'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，验证集损失：{self.valid_loss_values[-1]}")
-                    else:
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}")
+                        text += f"，验证集损失：{self.valid_loss_values[-1]}"
+                    print(text)
 
-            return self.loss_values, self.valid_loss_values
+                if isinstance(x_valid, list) and isinstance(y_valid, list) and self.validation_set_early_stop is not False:
+                    if self.valid_loss_values[-1] < self.min_valid_loss:
+                        self.min_valid_cnt = 0
+                        self.min_valid_loss = self.valid_loss_values[-1]
+                    else:
+                        self.min_valid_cnt += 1
+
+                if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
+                    break
+
+
+            if isinstance(x_valid, list) and isinstance(y_valid, list):
+                return self.loss_values, self.valid_loss_values
+            else:
+                return self.loss_values
 
         if self.target == 'output_multi_node_regress':
             for _ in tqdm(range(epochs)):
+                text = f'第{_ + 1}代'
+                if isinstance(methods, dict):
+                    if 'lr_decay' in methods:
+                        lr = lr0 * np.exp(-methods['lr_decay'][1] * _)
+                        text += f'学习率：{lr}，'
 
                 indices = list(range(len(x_train)))
                 random.shuffle(indices)
@@ -587,12 +673,26 @@ class MLP:
                     self.valid_loss_values.append(valid_epoch_loss / len(x_valid))
 
                 if (_ + 1) % verbose_frequence == 0:
+                    text += f'损失：{self.loss_values[-1]}'
                     if isinstance(x_valid, list) and isinstance(y_valid, list):
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}，验证集损失：{self.valid_loss_values[-1]}")
-                    else:
-                        print(f"第{_ + 1}代损失：{self.loss_values[-1]}")
+                        text += f"，验证集损失：{self.valid_loss_values[-1]}"
+                    print(text)
 
-            return self.loss_values, self.valid_loss_values
+                if isinstance(x_valid, list) and isinstance(y_valid, list) and self.validation_set_early_stop is not False:
+                    if self.valid_loss_values[-1] < self.min_valid_loss:
+                        self.min_valid_cnt = 0
+                        self.min_valid_loss = self.valid_loss_values[-1]
+                    else:
+                        self.min_valid_cnt += 1
+
+                if self.min_valid_cnt == self.validation_set_early_stop and self.validation_set_early_stop is not False:
+                    print(f'验证集损失{self.validation_set_early_stop}代不减小，提前终止')
+                    break
+
+            if isinstance(x_valid, list) and isinstance(y_valid, list):
+                return self.loss_values, self.valid_loss_values
+            else:
+                return self.loss_values
 
     def predict(self, x):
         if self.target == '2_classify':
